"""
Implementation of AI classes
"""
import dataclasses
import datetime
import json
from enum import Enum
from typing import Type, Optional

import tiktoken

from merlin_ai.data_classes import NativeDataClass
from merlin_ai.llm_classes import PromptBase, OpenAIPrompt
from merlin_ai.settings import default_model_settings


class BaseAIClass:
    """
    Base class for AI classes
    """

    def __init__(self, data_class: Type, model_settings: Optional[dict] = None):
        self._data_class = data_class
        self._model_settings = model_settings

    def as_prompt(
        self,
        value: str,
        model_settings: Optional[dict] = None,
        instruction: Optional[str] = None,
    ) -> dict:
        """
        Get prompt which would be used when evaluating model.
        """
        llm_settings = self._get_llm_settings(model_settings)
        prompt_dict = self._generate_prompt(
            value, model_settings=llm_settings, instruction=instruction
        ).as_dict()

        if "api_key" in prompt_dict:
            del prompt_dict["api_key"]

        return prompt_dict

    def _generate_prompt(
        self, value: str, model_settings: dict, instruction: Optional[str] = None
    ) -> PromptBase:
        """
        Generate LLM prompt from provided value
        :param value:
        :return:
        """
        raise NotImplementedError()

    def _get_llm_settings(
        self, function_call_model_settings: Optional[dict] = None
    ) -> dict:
        """
        Get LLM settings
        """
        settings = default_model_settings
        if self._model_settings:
            settings.update(self._model_settings)
        if function_call_model_settings:
            settings.update(function_call_model_settings)

        return {key: value for key, value in settings.items() if value is not None}

    def _create_instance_from_response(self, llm_response):
        """
        Create base class instance from LLM response
        """
        raise NotImplementedError()

    def __call__(
        self,
        value: str,
        model_settings: Optional[dict] = None,
        instruction: Optional[str] = None,
        return_raw_response: bool = False,
    ):
        """
        Call LLM model, parse the response and instantiate data class
        :param value: user provided string from which to instantiate data class
        :param model_settings: settings for LLM model
        :param instruction: custom instruction to use when generating LLM prompt
        :param return_raw_response: if True raw response from LLM API will be returned alongside the AIModel instance
        :return: instance of data class
        """
        llm_settings = self._get_llm_settings(model_settings)

        prompt = self._generate_prompt(
            value, model_settings=llm_settings, instruction=instruction
        )
        response = prompt.get_llm_response()

        instance = self._create_instance_from_response(response)
        if return_raw_response:
            return instance, response

        return instance


class OpenAIModel(BaseAIClass):
    """
    OpenAI-based AI Model
    """

    def __str__(self):
        return f"OpenAIModel: {self._data_class.__name__}"

    @staticmethod
    def _generate_function_call_object(data_class: Type) -> dict:
        """
        Generate function call object from data class
        """
        if dataclasses.is_dataclass(data_class):
            return NativeDataClass(data_class).generate_format_function_call_object()

        raise ValueError(f"Unsupported data class: {data_class}")

    def _create_instance_from_response(self, llm_response):
        function_call_response = llm_response.choices[0].message.function_call
        function_name = "format_response"
        if not function_call_response or function_call_response.name != function_name:
            raise RuntimeError(f"LLM did not call function '{function_name}'")

        arguments = json.loads(function_call_response.arguments)
        return self._data_class(**arguments)

    def _generate_prompt(
        self,
        value: str,
        model_settings: Optional[dict] = None,
        instruction: Optional[str] = None,
    ) -> PromptBase:
        function_name = "format_response"
        model_settings["functions"] = [
            self._generate_function_call_object(self._data_class)
        ]
        model_settings["function_call"] = {"name": function_name}
        if not instruction:
            instruction = self._data_class.__doc__.strip()

        system_instruction = ""
        if not instruction.startswith(f"{self._data_class.__name__}("):
            system_instruction = f"Also note that: {instruction}\n\n"

        return OpenAIPrompt(
            model_settings,
            messages=[
                {
                    "role": "system",
                    "content": "The user will provide text that you need to parse into a structured form.\n"
                    f"To validate your response, you must call the `{function_name}` function.\n"
                    "Use the provided text and context to extract, deduce, or infer\n"
                    f"any parameters needed by `{function_name}`, including any missing data.\n\n{system_instruction}"
                    "You have been provided the following context to perform your task:\n"
                    f"    - The current time is {datetime.datetime.now()}.",
                },
                {"role": "user", "content": f"The text to parse:\n{value}"},
            ],
        )


class OpenAIEnum(BaseAIClass):
    """
    OpenAI-based AI Enum
    """

    def __str__(self):
        return f"OpenAIEnum: {self._data_class.__name__}"

    def _create_instance_from_response(self, llm_response):
        content = llm_response.choices[0].message.content
        enum_options = self._get_enum_options()

        try:
            enum_option_index = int(content)
            if enum_option_index < 1 or enum_option_index > len(enum_options):
                raise ValueError()
        except (TypeError, ValueError) as err:
            raise RuntimeError(f"LLM returned invalid value {content}") from err

        return enum_options[int(content) - 1]

    def _get_llm_settings(
        self, function_call_model_settings: Optional[dict] = None
    ) -> dict:
        """
        Get LLM settings
        """
        settings = super()._get_llm_settings(function_call_model_settings)
        settings["max_tokens"] = 1
        return settings

    def _get_enum_options(self) -> list:
        """
        Get enum options.
        """
        if not issubclass(self._data_class, Enum):
            raise ValueError(f"{self._data_class} is not an Enum")

        return list(self._data_class)

    def _generate_prompt(
        self, value: str, model_settings: dict, instruction: Optional[str] = None
    ) -> PromptBase:
        enum_options = self._get_enum_options()
        encoding = tiktoken.get_encoding("cl100k_base")
        model_settings["logit_bias"] = {
            str(encoding.encode(str(idx))[0]): 100
            for idx in range(1, len(enum_options) + 1)
        }
        system_prompt = "You are an expert classifier that always chooses correctly\n\n"

        if not instruction:
            instruction = self._data_class.__doc__.strip()
        if instruction != "An enumeration.":
            system_prompt += f"Also note that:\n{instruction}\n\n"

        system_prompt += (
            "The user will provide text to classify, you will use your expertise "
            "to choose the best option below based on it:\n"
            + "\n".join(
                [
                    f"\t{idx+1}. {option.name} ({idx+1})"
                    for idx, option in enumerate(enum_options)
                ]
            )
        )

        return OpenAIPrompt(
            model_settings,
            messages=[
                {
                    "role": "system",
                    "content": system_prompt,
                },
                {"role": "user", "content": f"The text to classify:\n{value}"},
            ],
        )



"""
Decorators for AI classes
"""
from typing import Type, Optional, Callable, Union

from merlin_ai.ai_classes import OpenAIModel, BaseAIClass, OpenAIEnum


def decorator_generator(ai_class: Type[BaseAIClass]):
    """
    Generator for decorators
    :param ai_class: AI Class type for which to  generate a decorator
    """

    def ai_class_decorator(*args, **kwargs) -> Union[BaseAIClass, Callable]:
        """
        Turn base class into AI Model
        """

        def _create_ai_model(model_settings: Optional[dict] = None):
            """
            Create AI Model
            :param model_settings:
            :return:
            """

            def _inner(base_class: Type):
                """
                Inner function
                :param base_class:
                :return:
                """
                return ai_class(base_class, model_settings=model_settings)

            return _inner

        if len(args) == 1 and callable(args[0]):
            return _create_ai_model(kwargs)(args[0])

        return _create_ai_model(kwargs)

    return ai_class_decorator


ai_model = decorator_generator(OpenAIModel)
ai_enum = decorator_generator(OpenAIEnum)



"""
Supported data classes as a base of AI Models
"""
import dataclasses
import datetime
import typing
from typing import Type


class BaseSupportedDataClass:
    """
    Base class
    """

    def __init__(self, data_type):
        self.data_type = data_type

    def generate_format_function_call_object(self) -> dict:
        """
        Generate function call object for OpenAI
        """
        raise NotImplementedError()


class NativeDataClass(BaseSupportedDataClass):
    """
    Support for python's native Data
    """

    def __init__(self, data_type):
        super().__init__(data_type)
        if not dataclasses.is_dataclass(data_type):
            raise ValueError(f"{data_type} is not dataclass")

    @classmethod
    def _get_type_args(cls, param_type: Type) -> dict:
        """
        Parameter type to args
        """
        if param_type == str:
            return {"type": "string"}

        if param_type == datetime.date:
            return {"type": "string", "format": "date"}

        if param_type == datetime.datetime:
            return {"type": "string", "format": "date"}

        if param_type == datetime.time:
            return {"type": "string", "format": "time"}

        if param_type == datetime.timedelta:
            return {"type": "string", "format": "duration"}

        if typing.get_origin(param_type) == typing.Optional:
            base_type_args = cls._get_type_args(typing.get_args(param_type)[0])
            base_type = base_type_args["type"]
            if isinstance(base_type, list):
                base_type.append("null")
            if isinstance(base_type, str):
                base_type_args["type"] = [base_type, "null"]

            return base_type_args

        if typing.get_origin(param_type) == list:
            generic_type = None
            if typing.get_args(param_type):
                generic_type = cls._get_type_args(typing.get_args(param_type)[0])
            return {
                "type": "array",
                "items": generic_type or {"type": "string"},
            }
        if typing.get_origin(param_type) == typing.Literal:
            options = typing.get_args(param_type)
            type_args = cls._get_type_args(type(options[0]))
            type_args["enum"] = list(options)
            return type_args

        return {}

    @classmethod
    def _generate_parameters(cls, data_class) -> dict:
        """
        Generate parameters argument
        """
        properties = {}

        for field in dataclasses.fields(data_class):
            field_property = {"title": field.name.replace("_", " ").capitalize()}
            field_property.update(cls._get_type_args(field.type))

            description = field.metadata.get("description")
            if description:
                field_property["description"] = description

            properties[field.name] = field_property

        return {
            "properties": properties,
            "required": list(properties.keys()),
            "type": "object",
        }

    def generate_format_function_call_object(self):
        function_call_obj = {
            "name": "format_response",
            "parameters": self._generate_parameters(self.data_type),
        }

        function_description: str = self.data_type.__doc__
        if function_description and not function_description.startswith(
            f"{self.data_type.__name__}("
        ):
            function_call_obj["description"] = function_description.strip()

        return function_call_obj



"""
LLM classes
"""
import os
from typing import Optional

import openai


class PromptBase:
    """
    Base class for creating Prompts
    """

    def as_dict(self) -> dict:
        """
        Get prompt as kwargs dict for library call
        """
        raise NotImplementedError()

    def get_llm_response(self):
        """
        Get response to the prompt from LLM
        """
        raise NotImplementedError()


class OpenAIPrompt(PromptBase):
    """
    Prompt for OpenAI models
    """

    def __init__(
        self,
        model_settings: dict,
        messages: Optional[list[dict]] = None,
        system_message: Optional[str] = None,
    ):
        self._model_settings = model_settings
        self._messages = messages or []

        if system_message:
            self.add_message("system", system_message)

    def add_message(self, role: str, content: str):
        """
        Add message to prompt messages
        :param role: role of the message
        :param content: Content of the message
        """
        self._messages.append(({"role": role, "content": content}))

    def as_dict(self) -> dict:
        return {
            **self._model_settings,
            "messages": self._messages,
        }

    def get_llm_response(self):
        openai.api_key = os.environ.get("OPENAI_API_KEY")
        return openai.ChatCompletion.create(**self.as_dict())
